@@ -22,7 +22,9 @@ import cups
 import cupspk
 import gobject
 import gtk
+import os
 from errordialogs import *
+import config
 from debug import *
 
 _ = lambda x: x
@@ -130,11 +132,11 @@ class Connection:
         self._server = cups.getServer ()
         self._port = cups.getPort()
         self._encryption = cups.getEncryption ()
-        self._connect ()
         self._prompt_allowed = True
         self._operation_stack = []
         self._lock = lock
         self._gui_event = threading.Event ()
+        self._connect ()
 
     def _begin_operation (self, operation):
         self._operation_stack.append (operation)
@@ -154,7 +156,12 @@ class Connection:
     def _connect (self):
         cups.setUser (self._use_user)
 
-        self._use_pk = self._server[0] == '/' or self._server == 'localhost'
+        self._use_pk = ((self._server[0] == '/' or self._server == 'localhost')
+                        and os.getuid () != 0)
+
+        if not config.WITH_POLKIT_1 and self._lock:
+            self._use_pk = False
+
         if self._use_pk:
             create_object = cupspk.Connection
         else:
@@ -205,10 +212,6 @@ class Connection:
                 break
             except cups.IPPError, (e, m):
                 if self._use_pk and m == 'pkcancel':
-                    title = _('Unauthorized request (%s)') % fname
-                    text = _("You are not authorized to carry out the "
-                             "requested action.")
-                    show_error_dialog (title, text, None)
                     raise cups.IPPError (0, _("Operation canceled"))
                 if not self._cancel and (e == cups.IPP_NOT_AUTHORIZED or
                                          e == cups.IPP_FORBIDDEN):
@@ -388,6 +391,8 @@ class Connection:
                       _("The password may be incorrect."))
         if self._lock:
             d.connect ("response", self._on_not_authorized_dialog_response)
+            d.show_all ()
+            d.show_now ()
             gtk.gdk.threads_leave ()
         else:
             d.run ()
