@@ -89,6 +89,7 @@ class Printer:
         if self.is_shared is None:
             self.is_shared = not self.not_shared
         del self.not_shared
+        self.other_attributes = kw
 
     def getAttributes(self):
         """
@@ -156,12 +157,6 @@ class Printer:
         elif attrs.has_key('requesting-user-name-denied'):
             self.except_users = attrs['requesting-user-name-denied']
         self.except_users_string = ', '.join(self.except_users)
-        if (attrs.has_key ('device-uri') and
-            attrs['device-uri'].startswith ("smb:")):
-            # Don't update device-uri for smb printers as we need to have
-            # the original from printers.conf, in case it has authentication
-            # details in it.
-            attrs['device-uri'] = self.device_uri
         self.update (**attrs)
 
     def getServer(self):
@@ -322,7 +317,7 @@ class Printer:
         else:
             self.connection.setPrinterUsersAllowed(self.name, except_users)
 
-    def jobsQueued(self, only_tests=False):
+    def jobsQueued(self, only_tests=False, limit=None):
         """
         Find out whether jobs are queued for this printer.
 
@@ -349,15 +344,47 @@ class Printer:
                 (attrs.has_key ('job-name') and
                  attrs['job-name'] == 'Test Page')):
                 ret.append (id)
+
+                if limit != None and len (ret) == limit:
+                    break
         return ret
 
-    def testsQueued(self):
+    def jobsPreserved(self, limit=None):
+        """
+        Find out whether there are preserved jobs for this printer.
+
+        @return: list of job IDs
+        """
+        ret = []
+        try:
+            jobs = self.connection.getJobs (which_jobs='completed')
+        except cups.IPPError:
+            return ret
+
+        for id, attrs in jobs.iteritems():
+            try:
+                uri = attrs['job-printer-uri']
+                uri = uri[uri.rindex ('/') + 1:]
+            except:
+                continue
+            if uri != self.name:
+                continue
+            if (attrs.get ('job-state',
+                           cups.IPP_JOB_PENDING) < cups.IPP_JOB_COMPLETED):
+                continue
+            ret.append (id)
+            if limit != None and len (ret) == limit:
+                break
+
+        return ret
+
+    def testsQueued(self, limit=None):
         """
         Find out whether test jobs are queued for this printer.
 
         @returns: list of job IDs
         """
-        return self.jobsQueued (only_tests=True)
+        return self.jobsQueued (only_tests=True, limit=limit)
 
     def setAsDefault(self):
         """
@@ -514,7 +541,7 @@ class Device:
         
         return result
 
-def getDevices(connection):
+def getDevices(connection, **kw):
     """
     Obtain a list of available CUPS devices.
 
@@ -523,7 +550,7 @@ def getDevices(connection):
     @returns: a list of L{Device} objects
     @raise cups.IPPError: IPP Error
     """
-    devices = connection.getDevices()
+    devices = connection.getDevices(**kw)
     for uri, data in devices.iteritems():
         device = Device(uri, **data)
         devices[uri] = device
@@ -616,6 +643,10 @@ def missingPackagesAndExecutables(ppd):
     # First, a local function.  How to check that something exists
     # in a path:
     def pathcheck (name, path="/usr/bin:/bin"):
+        if name == "-":
+            # A filter of "-" means that no filter is required,
+            # i.e. the device accepts the given format as-is.
+            return "builtin"
         # Strip out foomatic '%'-style place-holders.
         p = name.find ('%')
         if p != -1:
@@ -742,8 +773,10 @@ def missingPackagesAndExecutables(ppd):
             # IJS servers (used by foomatic)
             'hpijs': 'hpijs',
             'ijsgutenprint.5.0': 'gutenprint',
+            'ijsgutenprint.5.2': 'gutenprint',
             # CUPS filters
             'rastertogutenprint.5.0': 'gutenprint-cups',
+            'rastertogutenprint.5.2': 'gutenprint-cups',
             'commandtoepson': 'gutenprint-cups',
             'commandtocanon': 'gutenprint-cups',
             }
