@@ -26,7 +26,7 @@ import string
 import locale
 import os.path
 import re
-from . import _debugprint
+from . import _debugprint, set_debugprint_fn
 
 __all__ = ['ppdMakeModelSplit',
            'PPDs']
@@ -43,61 +43,84 @@ def ppdMakeModelSplit (ppd_make_and_model):
     # If the string starts with a known model name (like "LaserJet") assume
     # that the manufacturer name is missing and add the manufacturer name
     # corresponding to the model name
-    if re.search ("^\s*(deskjet|dj\b|dj\d|laserjet|lj\b|color\s*laserjet|color\s*lj\b|designjet|officejet|oj\b|photosmart|ps\b|psc)", \
-                      ppd_make_and_model, re.I) or \
-       re.search ("(edgeline)", ppd_make_and_model, re.I):
+    ppd_make_and_model.strip ()
+    l = ppd_make_and_model.lower ()
+    if (l.startswith ("deskjet") or
+        l.startswith ("dj ") or l == "dj" or
+        (l.startswith ("dj") and len (l) > 2 and l[2].isdigit ()) or
+        l.startswith ("laserjet") or
+        l.startswith ("lj") or
+        l.startswith ("color laserjet") or
+        l.startswith ("color lj") or
+        l.startswith ("designjet") or
+        l.startswith ("officejet") or
+        l.startswith ("oj") or
+        l.startswith ("photosmart") or
+        l.startswith ("ps ") or
+        l.startswith ("psc") or
+        l.startswith ("edgeline")):
         make = "HP"
         model = ppd_make_and_model
-    elif re.search ("^\s*(stylus|aculaser)", \
-                      ppd_make_and_model, re.I):
+    elif (l.startswith ("stylus") or
+          l.startswith ("aculaser")):
         make = "Epson"
         model = ppd_make_and_model
-    elif re.search ("^\s*(stylewriter|imagewriter|deskwriter|laserwriter)", \
-                      ppd_make_and_model, re.I):
+    elif (l.startswith ("stylewriter") or
+          l.startswith ("imagewriter") or
+          l.startswith ("deskwriter") or
+          l.startswith ("laserwriter")):
         make = "Apple"
         model = ppd_make_and_model
-    elif re.search ("^\s*(pixus|pixma|selphy|imagerunner|\bbjc\b|\bbj\b|\blbp\b)",\
-                      ppd_make_and_model, re.I):
+    elif (l.startswith ("pixus") or
+          l.startswith ("pixma") or
+          l.startswith ("selphy") or
+          l.startswith ("imagerunner") or
+          l.startswith ("bj") or
+          l.startswith ("lbp")):
         make = "Canon"
         model = ppd_make_and_model
-    elif re.search ("^\s*(\bhl\b|\bdcp\b|\bmfc\b)", \
-                      ppd_make_and_model, re.I):
+    elif (l.startswith ("hl") or
+          l.startswith ("dcp") or
+          l.startswith ("mfc")):
         make = "Brother"
         model = ppd_make_and_model
-    elif re.search ("^\s*(docuprint|docupage|phaser|workcentre|homecentre)", \
-                      ppd_make_and_model, re.I):
+    elif (l.startswith ("docuprint") or
+          l.startswith ("docupage") or
+          l.startswith ("phaser") or
+          l.startswith ("workcentre") or
+          l.startswith ("homecentre")):
         make = "Xerox"
         model = ppd_make_and_model
-    elif re.search ("^\s*(optra|(color\s*|)jetprinter)", \
-                      ppd_make_and_model, re.I):
+    elif (l.startswith ("optra") or
+          l.startswith ("jetprinter") or
+          l.startswith ("color jetprinter")):
         make = "Lexmark"
         model = ppd_make_and_model
-    elif re.search ("^\s*(magicolor|pageworks|pagepro)", \
-                      ppd_make_and_model, re.I):
+    elif (l.startswith ("magicolor") or
+          l.startswith ("pageworks") or
+          l.startswith ("pagepro")):
         make = "KONICA MINOLTA"
         model = ppd_make_and_model
-    elif re.search ("^\s*(aficio)", \
-                      ppd_make_and_model, re.I):
+    elif l.startswith ("aficio"):
         make = "Ricoh"
         model = ppd_make_and_model
-    elif re.search ("^\s*(varioprint)", \
-                      ppd_make_and_model, re.I):
+    elif l.startswith ("varioprint"):
         make = "Oce"
         model = ppd_make_and_model
-    elif re.search ("^\s*(okipage|microline)", \
-                      ppd_make_and_model, re.I):
+    elif (l.startswith ("okipage") or
+          l.startswith ("microline")):
         make = "Oki"
         model = ppd_make_and_model
-    elif re.search ("^\s*(konica[\s_-]*minolta)", \
-                      ppd_make_and_model, re.I):
-        make = "KONICA MINOLTA"
-        model = ppd_make_and_model
-        model = re.sub ("(?i)KONICA[\s_-]*MINOLTA\s*", "", model, 1)
-    elif re.search("turboprint", ppd_make_and_model, re.I):
-        # Support for Turboprint PPDs
+
+    # Handle PPDs provided by Turboprint
+    elif l.find ("turboprint") != -1:
         t = ppd_make_and_model.find (" TurboPrint")
         if t != -1:
-            ppd_make_and_model = ppd_make_and_model[:t]
+            t2 = ppd_make_and_model.rfind (" TurboPrint")
+            if t != t2:
+                ppd_make_and_model = ppd_make_and_model[t + 12:t2]
+            else:
+                ppd_make_and_model = ppd_make_and_model[:t]
         try:
             make, model = ppd_make_and_model.split("_", 1)
         except:
@@ -109,6 +132,13 @@ def ppdMakeModelSplit (ppd_make_and_model):
         model = re.sub (r"(?<=[a-z])(?=[A-Z])", " ", model)
         model = re.sub (r" Jet", "Jet", model)
         model = re.sub (r"Photo Smart", "PhotoSmart", model)
+
+    # Special handling for two-word manufacturers
+    elif l.startswith ("konica minolta "):
+        make = "KONICA MINOLTA"
+        model = ppd_make_and_model[15:]
+
+    # Finally, take the first word as the name of the manufacturer.
     else:
         try:
             make, model = ppd_make_and_model.split(" ", 1)
@@ -116,70 +146,65 @@ def ppdMakeModelSplit (ppd_make_and_model):
             make = ppd_make_and_model
             model = ''
 
-    # Model names do not contain a comma, truncate all from the
-    # comma on
-    c = model.find (",")
-    if c != -1:
-        model = model[:c]
+    # Standardised names for manufacturers.
+    makel = make.lower ()
+    if (makel.startswith ("konica") and
+        makel.endswith ("minolta")):
+        make = "KONICA MINOLTA"
+        makel = "konica minolta"
+    elif (makel.startswith ("hewlett") and
+          makel.endswith ("packard")):
+        make = "HP"
+        makel = "hp"
+    elif makel == "lexmark international":
+        make = "Lexmark"
+        makel = "lexmark"
 
     # HP PPDs give NickNames like:
     # *NickName: "HP LaserJet 4 Plus v2013.111 Postscript (recommended)"
     # Find the version number.
-    v = model.find (" v")
+    modell = model.lower ()
+    v = modell.find (" v")
     if v != -1 and (model[v + 2].isdigit () or
                     (model[v + 2] == '.' and
                      model[v + 3].isdigit ())):
         # Truncate at that point.
         model = model[:v]
+        modell = modell[:v]
 
-    h = model.find (" hpijs")
-    if h != -1:
-        model = model[:h]
+    for suffix in [" hpijs",
+                   " foomatic/",
+                   " - ",
+                   " w/",
+                   " (",
+                   " postscript",
+                   " ps",
+                   " ps1",
+                   " ps2",
+                   " ps3",
+                   " pxl",
+                   " series",
+                   "_bt",
+                   ","]:
+        s = modell.find (suffix)
+        if s != -1:
+            model = model[:s]
+            modell = modell[:s]
 
-    f = model.find (" Foomatic/")
-    if f != -1:
-        model = model[:f]
-
-    # Gutenprint PPDs have NickNames that end:
-    # ... - CUPS+Gutenprint v5.0.0
-    gutenprint = model.find (" - CUPS+Gutenprint")
-    if gutenprint != -1:
-        model = model[:gutenprint]
-
-    # Gimp-Print PPDs have NickNames that end:
-    # ... - CUPS+Gimp-Print v4.2.7
-    gimpprint = model.find (" - CUPS+Gimp-Print")
-    if gimpprint != -1:
-        model = model[:gimpprint]
-
-    wth = model.find (" w/")
-    if wth != -1:
-        model = model[:wth]
-
-    make = re.sub (r"(?i)KONICA[\s_-]*MINOLTA", "KONICA MINOLTA", make, 1)
-    make = re.sub (r"(?i)HEWLETT[\s_-]PACKARD", "HP", make, 1)
-    make = re.sub (r"(?i)Lexmark\s*International", "Lexmark", make, 1)
-    model = re.sub (r"(?i)\s*\(recommended\)", "", model)
-    model = re.sub (r"(?i)\s*-\s*PostScript\b", "", model)
-    model = re.sub (r"(?i)\s*\bseries\b", "", model)
-    if make.lower () == "hp":
-        model = re.sub (r"(?i)^dj\b", "DeskJet", model)
-        model = re.sub (r"(?i)^LJ\b", "LaserJet", model)
-        model = re.sub (r"(?i)^OJ\b", "OfficeJet", model)
-        model = re.sub (r"(?i)^Color\s*LJ\b", "Color LaserJet", model)
-        model = re.sub (r"(?i)^PS\b", "PhotoSmart", model)
-    model = re.sub (r"(?i)\s*\bPS[123]?\b", "", model)
-    model = re.sub (r"(?i)\s*\bPXL", "", model)
-    model = re.sub (r"(?i)[\s_-]+BT\b", "", model)
-    model = re.sub (r"(?i)\s*\(Bluetooth\)", "", model)
-    model = re.sub (r"(?i)\s*-\s*(RC|Ver(|sion))\s*-*\s*[0-9\.]+", "", model)
-    model = re.sub (r"(?i)\s*-\s*(RC|Ver(|sion))\b", "", model)
-    model = re.sub (r"(?i)\s*PostScript\s*$", "", model)
-    model = re.sub (r"(?i)\s*\(\s*\)", "", model)
-    model = re.sub (r"(?i)\s*[\-\/]\s*$", "", model)
+    if makel == "hp":
+        modelnames = {"dj": "DeskJet",
+                      "lj": "LaserJet",
+                      "oj": "OfficeJet",
+                      "color lj": "Color LaserJet",
+                      "ps ": "PhotoSmart",
+                      "hp ": ""}
+        for (name, fullname) in modelnames.iteritems ():
+            if modell.startswith (name):
+                model = fullname + model[len (name):]
+                modell = model.lower ()
 
     for mfr in [ "Apple", "Canon", "Epson", "Lexmark", "Oki" ]:
-        if make == mfr.upper ():
+        if makel == mfr.lower ():
             make = mfr
 
     model = model.strip ()
@@ -191,6 +216,7 @@ DRIVER_TYPE_DOWNLOADED_NOW = 5
 DRIVER_TYPE_FOOMATIC_RECOMMENDED_NON_POSTSCRIPT = 8
 DRIVER_TYPE_VENDOR = 10
 DRIVER_TYPE_FOOMATIC_RECOMMENDED_POSTSCRIPT = 15
+DRIVER_TYPE_HPCUPS = 16
 DRIVER_TYPE_FOOMATIC_HPIJS_ON_HP = 17
 DRIVER_TYPE_GUTENPRINT_NATIVE_SIMPLIFIED = 20
 DRIVER_TYPE_GUTENPRINT_NATIVE = 25
@@ -208,14 +234,10 @@ def _getDriverType (ppdname, ppds=None):
     """Decides which of the above types ppdname is."""
     if ppdname.find ("turboprint") != -1:
         return DRIVER_TYPE_3RD_PARTY_NONFREE
-    if ppdname.find ("gutenprint") != -1:
-        if (ppdname.find ("/simple/") != -1 or
-            ppdname.find (".sim-") != -1):
-            return DRIVER_TYPE_GUTENPRINT_NATIVE_SIMPLIFIED
-        else:
-            return DRIVER_TYPE_GUTENPRINT_NATIVE
     if ppdname.find ("splix")!= -1:
         return DRIVER_TYPE_SPLIX
+    if ppdname.find ("hpcups") != -1:
+        return DRIVER_TYPE_HPCUPS
     if (ppdname.find (":") == -1 and
         ppdname.find ("/cups-included/") != -1):
         return DRIVER_TYPE_CUPS
@@ -241,6 +263,12 @@ def _getDriverType (ppdname, ppds=None):
                 return DRIVER_TYPE_FOOMATIC_GUTENPRINT_SIMPLIFIED
             return DRIVER_TYPE_FOOMATIC_GUTENPRINT
         return DRIVER_TYPE_FOOMATIC
+    if ppdname.find ("gutenprint") != -1:
+        if (ppdname.find ("/simple") != -1 or
+            ppdname.find (".sim-") != -1):
+            return DRIVER_TYPE_GUTENPRINT_NATIVE_SIMPLIFIED
+        else:
+            return DRIVER_TYPE_GUTENPRINT_NATIVE
     if ppdname.find ("-hpijs") != -1:
         if ppdname.find ("hpijs-rss") == -1:
             return DRIVER_TYPE_FOOMATIC_HPIJS
@@ -344,7 +372,7 @@ class PPDs:
             models_list = self.makes[make].keys ()
         except KeyError:
             return []
-        models_list.sort (key=unicode.lower, cmp=cups.modelSort)
+        models_list.sort (key=lambda x: x.lower(), cmp=cups.modelSort)
         return models_list
 
     def getInfoFromModel (self, make, model):
@@ -392,16 +420,10 @@ class PPDs:
                 (xpath, xslash, xfilename) = x.rpartition ("/")
                 if filename == xfilename:
                     return DRIVER_TYPE_DOWNLOADED_NOW
-            if mfg == "HP":
+            if mfg == "HP" or mfg == "Apollo":
                 if t == DRIVER_TYPE_FOOMATIC_HPIJS:
                     # Prefer HPIJS for HP devices.
                     t = DRIVER_TYPE_FOOMATIC_HPIJS_ON_HP
-                    # For HP LaserJet 12xx/13xx prefer HPIJS over
-                    # PostScript, as they do not have enough memory
-                    # to render complex graphics with their on-board
-                    # PostScript interpreter
-                    if re.search(r"(?i)HP[-_]LaserJet_1[23]\d\d", x):
-                        t = DRIVER_TYPE_FOOMATIC_RECOMMENDED_NON_POSTSCRIPT
             return t
 
         def sort_ppdnames (a, b):
@@ -502,90 +524,80 @@ class PPDs:
         _debugprint ("\n%s %s" % (mfg, mdl))
         self._init_ids ()
 
-        # Consider "HP" and "Hewlett-Packard" as equal, as the ID returned
-        # by the CUPS "usb" backend and HPLIP's "hp" backend are different
-        if mfg.lower () == "hewlett-packard":
-            mfg = "HP"
-
+        # Start with an empty result list and build it up using
+        # several search methods, in increasing order of fuzziness.
         ppdnamelist = []
+
+        # First, try looking up the device using the manufacturer and
+        # model fields from the Device ID exactly as they appear (but
+        # case-insensitively).
+        mfgl = mfg.lower ()
+        mdll = mdl.lower ()
+
         id_matched = False
         try:
-            ppdnamelist = self.ids[mfg.lower ()][mdl.lower ()]
+            ppdnamelist = self.ids[mfgl][mdll]
             status = self.STATUS_SUCCESS
             id_matched = True
         except KeyError:
             pass
 
-        mfgl = mfg.lower ()
-        mdll = mdl.lower ()
-
-        # The HPLIP PPDs have incorrect IDs
-        for mf in ["hp", "apollo"]:
-            if mfgl == mf:
-                try:
-                    ppdnamelist += self.ids[mf][mdll]
-                    status = self.STATUS_SUCCESS
-                    id_matched = True
-                except KeyError:
-                    pass
-
-                try:
-                    ppdnamelist += self.ids[mf][mdll.replace (" ", "_")]
-                    status = self.STATUS_SUCCESS
-                    id_matched = True
-                except KeyError:
-                    pass
+        # The HP PPDs say "HP" not "Hewlett-Packard", so try that.
+        if mfgl == "hewlett-packard":
+            try:
+                ppdnamelist += self.ids["hp"][mdll]
+                status = self.STATUS_SUCCESS
+                id_matched = True
+            except KeyError:
+                pass
 
+        # Now try looking up the device by ppd-make-and-model.
         _debugprint ("Trying make/model names")
-        mfgl = mfg.lower ()
         mdls = None
         self._init_makes ()
-        for attempt in range (2):
-            for (make, models) in self.makes.iteritems ():
-                if make.lower () == mfgl:
-                    mdls = models
-                    break
+        make = None
+        if mfgl == "":
+            (mfg, mdl) = ppdMakeModelSplit (mdl)
+            mfgl = mfg.lower ()
+            mdll = mdl.lower ()
 
-            # Try again with replacements.
-            if mfgl == "hewlett-packard":
-                mfgl = "hp"
-            elif mfgl == "lexmark international":
-                mfgl = "lexmark"
-            elif mfgl == "":
-                (tmp1, tmp2) = ppdMakeModelSplit (mdl)
-                mfgl = tmp1.lower ()
+        mfgrepl = {"hewlett-packard": "hp",
+                   "lexmark international": "lexmark"}
+        if self.lmakes.has_key (mfgl):
+            # Found manufacturer.
+            make = self.lmakes[mfgl]
+        elif mfgrepl.has_key (mfgl):
+            rmfg = mfgrepl[mfgl]
+            if self.lmakes.has_key (rmfg):
+                mfg = rmfg
+                mfgl = mfg
+                # Found manufacturer (after mapping to canonical name)
+                make = self.lmakes[mfgl]
 
-        # Remove manufacturer name from model field
-        ppdnamelist2 = None
-        if mdl.startswith (mfg + ' '):
-            mdl = mdl[len (mfg) + 1:]
-        if mdl.startswith ('Hewlett-Packard '):
-            mdl = mdl[16:]
-        if mdl.startswith ('HP '):
-            mdl = mdl[3:]
-        if mdls:
-            for (model, ppdnames) in mdls.iteritems ():
-                if model.lower () == mdll:
-                    ppdnamelist2 = ppdnames.keys ()
-                    status = self.STATUS_SUCCESS
-                    break
-            if not ppdnamelist2:
+        if make != None:
+            mdls = self.makes[make]
+            mdlsl = self.lmodels[make.lower ()]
+
+            # Remove manufacturer name from model field
+            for prefix in [mfgl, 'hewlett-packard', 'hp']:
+                if mdll.startswith (prefix + ' '):
+                    mdl = mdl[len (prefix) + 1:]
+                    mdll = mdl.lower ()
+
+            if self.lmodels[mfgl].has_key (mdll):
+                model = mdlsl[mdll]
+                ppdnamelist += mdls[model].keys ()
+                status = self.STATUS_SUCCESS
+            else:
                 # Make use of the model name clean-up in the
                 # ppdMakeModelSplit () function
                 (mfg2, mdl2) = ppdMakeModelSplit (mfg + " " + mdl)
                 mdl2l = mdl2.lower ()
-                for (model, ppdnames) in mdls.iteritems ():
-                    if model.lower () == mdl2l:
-                        ppdnamelist2 = ppdnames.keys ()
-                        status = self.STATUS_SUCCESS
-                        break
+                if self.lmodels[mfgl].has_key (mdl2l):
+                    model = mdlsl[mdl2l]
+                    ppdnamelist += mdls[model].keys ()
+                    status = self.STATUS_SUCCESS
       
-        if ppdnamelist:
-            if ppdnamelist2:
-                ppdnamelist = ppdnamelist + ppdnamelist2
-        elif ppdnamelist2:
-            ppdnamelist = ppdnamelist2
-
         if not ppdnamelist and mdls:
             (s, ppds) = self._findBestMatchPPDs (mdls, mdl)
             if s != self.STATUS_NO_DRIVER:
@@ -666,8 +678,8 @@ class PPDs:
                 cmd = ""
 
             print "  <commandset>%s</commandset>" % cmd
-            print "Using %s" % ppdnamelist[0]
 
+        print "Using %s (status: %d)" % (ppdnamelist[0], status)
         return (status, ppdnamelist[0])
 
     def _findBestMatchPPDs (self, mdls, mdl):
@@ -926,7 +938,7 @@ def _self_test(argv):
 
     for opt, optarg in opts:
         if opt == "--help":
-            show_help ()
+            _show_help ()
             sys.exit (0)
         if opt == "--deviceid":
             stdin_deviceid = True
@@ -934,6 +946,14 @@ def _self_test(argv):
             list_models = True
         elif opt == "--list-ids":
             list_ids = True
+        elif opt == "--debug":
+            def _dprint(x):
+                try:
+                    print x
+                except:
+                    pass
+
+            set_debugprint_fn (_dprint)
 
     picklefile="pickled-ppds"
     import pickle
@@ -975,7 +995,7 @@ def _self_test(argv):
 
     idlist = [
         # Format is:
-        # (ID string, max status code, expected driver RE match)
+        # (ID string, max status code, expected ppd-make-and-model RE match)
 
         # Specific models
         ("MFG:EPSON;CMD:ESCPL2,BDC,D4,D4PX;MDL:Stylus D78;CLS:PRINTER;"
@@ -985,21 +1005,20 @@ def _self_test(argv):
         ("MFG:Hewlett-Packard;MDL:LaserJet 3390 Series;"
          "CMD:MLC,PCL,POSTSCRIPT;CLS:PRINTER;", 0, 'HP LaserJet 3390'),
         ("MFG:Hewlett-Packard;MDL:PSC 2200 Series;CMD:MLC,PCL,PML,DW-PCL,DYN;"
-         "CLS:PRINTER;1284.4DL:4d,4e,1;", 0, "HP PSC 2210"),
-        ("MFG:HP;MDL:PSC 2200 Series;CLS:PRINTER;DES:PSC 2200 Series;",
-         1, "HP PSC 2210"),# from HPLIP
+         "CLS:PRINTER;1284.4DL:4d,4e,1;", 0, "HP PSC 22[01]0"),
         ("MFG:HEWLETT-PACKARD;MDL:DESKJET 990C;CMD:MLC,PCL,PML;CLS:PRINTER;"
          "DES:Hewlett-Packard DeskJet 990C;", 0, "HP DeskJet 990C"),
         ("CLASS:PRINTER;MODEL:HP LaserJet 6MP;MANUFACTURER:Hewlett-Packard;"
          "DESCRIPTION:Hewlett-Packard LaserJet 6MP Printer;"
          "COMMAND SET:PJL,MLC,PCLXL,PCL,POSTSCRIPT;", 0, "HP LaserJet 6MP"),
+        # Canon PIXMA iP3000 (from gutenprint)
         ("MFG:Canon;CMD:BJL,BJRaster3,BSCCe;SOJ:TXT01;MDL:iP3000;CLS:PRINTER;"
          "DES:Canon iP3000;VER:1.09;STA:10;FSI:03;", 1, "Canon PIXMA iP3000"),
         ("MFG:HP;MDL:Deskjet 5400 series;CMD:MLC,PCL,PML,DW-PCL,DESKJET,DYN;"
          "1284.4DL:4d,4e,1;CLS:PRINTER;DES:5440;", 1, "HP DeskJet 5440"),
         ("MFG:Hewlett-Packard;MDL:HP LaserJet 3390;"
          "CMD:PJL,MLC,PCL,POSTSCRIPT,PCLXL;",
-         0, "HP LaserJet 3390.*Postscript"),
+         0, "HP LaserJet 3390"),
 
         # Generic models
         ("MFG:New;MDL:Unknown PS Printer;CMD:POSTSCRIPT;",
@@ -1023,21 +1042,23 @@ def _self_test(argv):
                                                          id_dict["MDL"],
                                                          id_dict["DES"],
                                                          id_dict["CMD"])
+        ppddict = ppds.getInfoFromPPDName (ppdname)
         if status < max_status_code:
             success = True
-        elif status == max_status_code:
-            ppddict = ppds.getInfoFromPPDName (ppdname)
-            match = re.match (modelre, ppddict['ppd-make-and-model'], re.I)
-            success = match != None
         else:
-            success = False
+            if status == max_status_code:
+                match = re.match (modelre, ppddict['ppd-make-and-model'], re.I)
+                success = match != None
+            else:
+                success = False
 
         if success:
             result = "PASS"
         else:
             result = "*** FAIL ***"
 
-        print "%s: %s %s" % (result, id_dict["MFG"], id_dict["MDL"])
+        print "%s: %s %s (%s)" % (result, id_dict["MFG"], id_dict["MDL"],
+                                  ppddict['ppd-make-and-model'])
         all_passed = all_passed and success
 
     if not all_passed:
