@@ -1,8 +1,8 @@
 # vim: set ts=4 sw=4 et: coding=UTF-8
 #
 # Copyright (C) 2008 Novell, Inc.
-# Copyright (C) 2009, 2009 Red Hat, Inc.
-# Copyright (C) 2009, 2009 Tim Waugh <twaugh@redhat.com>
+# Copyright (C) 2008, 2009 Red Hat, Inc.
+# Copyright (C) 2008, 2009 Tim Waugh <twaugh@redhat.com>
 #
 # Authors: Vincent Untz
 #
@@ -30,17 +30,18 @@ import os
 import sys
 
 import tempfile
-import string
 
 import cups
 import dbus
 import gtk
+import config
+from debug import debugprint
 
 from dbus.mainloop.glib import DBusGMainLoop
 DBusGMainLoop(set_as_default=True)
 
 PK_AUTH_NAME  = 'org.freedesktop.PolicyKit.AuthenticationAgent'
-PK_AUTH_PATH  = '/'
+PK_AUTH_PATH  = '/org/gnome/PolicyKit/Manager'
 PK_AUTH_IFACE = 'org.freedesktop.PolicyKit.AuthenticationAgent'
 
 CUPS_PK_NAME  = 'org.opensuse.CupsPkHelper.Mechanism'
@@ -151,7 +152,7 @@ class Connection:
         pk_auth_running = False
 
         if pk_auth_error != None:
-            if string.find(pk_auth_error, 'org.freedesktop.DBus.Error.NoReply') == 0:
+            if pk_auth_error.find('org.freedesktop.DBus.Error.NoReply') == 0:
                 return False
             raise dbus.exceptions.DBusException(pk_auth_error)
 
@@ -221,8 +222,14 @@ class Connection:
                             return retval
                 break
             except dbus.exceptions.DBusException, e:
-                if not self._handle_exception_with_auth(e):
+                if config.WITH_POLKIT_1:
+                    if e.get_dbus_name() == CUPS_PK_NEED_AUTH:
+                        raise cups.IPPError(cups.IPP_NOT_AUTHORIZED, 'pkcancel')
+
                     break
+                else:
+                    if not self._handle_exception_with_auth(e):
+                        break
 
         # The PolicyKit call did not work (either a PK-error and we got a dbus
         # exception that wasn't handled, or an error in the mechanism itself)
@@ -299,7 +306,86 @@ class Connection:
 #    getPPDs
 #    getServerPPD
 #    getDocument
-#    getDevices
+
+
+    def getDevices(self, *args, **kwds):
+        use_pycups = False
+
+        limit = 0
+        include_schemes = ''
+        exclude_schemes = ''
+        timeout = 0
+
+        if len(args) == 4:
+            (use_pycups, limit, include_schemes, exclude_schemes, timeout) = self._args_to_tuple([int, str, str, int], *args)
+        else:
+            if kwds.has_key('timeout'):
+                timeout = kwds['timeout']
+
+            if kwds.has_key('include_schemes'):
+                include_schemes = kwds['include_schemes']
+
+            if kwds.has_key('exclude_schemes'):
+                exclude_schemes = kwds['exclude_schemes']
+
+        # Convert from list to string
+        if len (include_schemes) > 0:
+            include_schemes = reduce (lambda x, y: x + "," + y, include_schemes)
+        else:
+            include_schemes = ""
+
+        if len (exclude_schemes) > 0:
+            exclude_schemes = reduce (lambda x, y: x + "," + y, exclude_schemes)
+        else:
+            exclude_schemes = ""
+
+        pk_args = (timeout, include_schemes, exclude_schemes)
+
+        result = self._call_with_pk_and_fallback(use_pycups,
+                                                 'DevicesGet', pk_args,
+                                                 self._connection.getDevices,
+                                                 *args, **kwds)
+
+        # return 'result' if fallback was called
+        if len (result.keys()) > 0 and type (result[result.keys()[0]]) == dict:
+             return result
+
+        result_str = {}
+        if result != None:
+            for i in result.keys():
+                if type(i) == dbus.String:
+                    result_str[str(i)] = str(result[i])
+                else:
+                    result_str[i] = result[i]
+
+        # cups-pk-helper returns all devices in one dictionary.
+        # Keys of different devices are distinguished by ':n' postfix.
+
+        devices = {}
+        n = 0
+        postfix = ':' + str (n)
+        device_keys = [x for x in result_str.keys() if x.endswith(postfix)]
+        while len (device_keys) > 0:
+
+            device_uri = None
+            device_dict = {}
+            for i in device_keys:
+                key = i[:len(i) - len(postfix)]
+                if key != 'device-uri':
+                    device_dict[key] = result_str[i]
+                else:
+                    device_uri = result_str[i]
+
+            if device_uri != None:
+                devices[device_uri] = device_dict
+
+            n += 1
+            postfix = ':' + str (n)
+            device_keys = [x for x in result_str.keys() if x.endswith(postfix)]
+
+        return devices
+
+
 #    getJobs
 #    getJobAttributes
 
@@ -680,7 +766,7 @@ class Connection:
 
         self._call_with_pk_and_fallback(use_pycups,
                                         'PrinterSetEnabled', pk_args,
-                                        self._connection.enablePrinter,
+                                        self._connection.disablePrinter,
                                         *args, **kwds)
 
 
@@ -718,8 +804,8 @@ class Connection:
         settings = {}
         if result != None:
             for i in result.keys():
-                if type(i) == 'dbus.String':
-                    settings[i.encode()] = result[i].encode()
+                if type(i) == dbus.String:
+                    settings[str(i)] = str(result[i])
                 else:
                     settings[i] = result[i]
 
