@@ -151,7 +151,6 @@ class JobViewer (GtkGUI, monitor.Watcher):
         self.job_creation_times_timer = None
         self.special_status_icon = False
         self.new_printer_notifications = {}
-        self.completed_job_notifications = {}
         self.authenticated_jobs = set() # of job IDs
 
         self.getWidgets ({"JobsWindow":
@@ -209,30 +208,44 @@ class JobViewer (GtkGUI, monitor.Watcher):
 
         self.job_menubar_item.set_submenu (self.job_context_menu)
 
-        text=0
-        for name in [_("Job"),
-                     _("User"),
-                     _("Document"),
-                     _("Printer"),
-                     _("Size"),
-                     _("Time submitted"),
-                     _("Status")]:
-            if text == 1 and trayicon:
+        for skip, ellipsize, name, setter in \
+                [(False, False, _("Job"), self._set_job_job_number_text),
+                 (True, False, _("User"), self._set_job_user_text),
+                 (False, True, _("Document"), self._set_job_document_text),
+                 (False, True, _("Printer"), self._set_job_printer_text),
+                 (False, False, _("Size"), self._set_job_size_text)]:
+            if trayicon and skip:
                 # Skip the user column for the trayicon.
-                text += 1
                 continue
+
             cell = gtk.CellRendererText()
-            if text == 2 or text == 3:
+            if ellipsize:
                 # Ellipsize the 'Document' and 'Printer' columns.
                 cell.set_property ("ellipsize", pango.ELLIPSIZE_END)
                 cell.set_property ("width-chars", 20)
-            column = gtk.TreeViewColumn(name, cell, text=text)
+            column = gtk.TreeViewColumn(name, cell)
+            column.set_cell_data_func (cell, setter)
             column.set_resizable(True)
             self.treeview.append_column(column)
-            text += 1
+
+        cell = gtk.CellRendererText ()
+        column = gtk.TreeViewColumn (_("Time submitted"), cell, text=1)
+        column.set_resizable (True)
+        self.treeview.append_column (column)
+
+        column = gtk.TreeViewColumn (_("Status"))
+        icon = gtk.CellRendererPixbuf ()
+        column.pack_start (icon, False)
+        text = gtk.CellRendererText ()
+        text.set_property ("ellipsize", pango.ELLIPSIZE_END)
+        text.set_property ("width-chars", 20)
+        column.pack_start (text, True)
+        column.set_cell_data_func (icon, self._set_job_status_icon)
+        column.set_cell_data_func (text, self._set_job_status_text)
+        self.treeview.append_column (column)
 
         self.treeview.get_selection().set_mode(gtk.SELECTION_SINGLE)
-        self.store = gtk.TreeStore(int, str, str, str, str, str, str)
+        self.store = gtk.TreeStore(int, str)
         self.store.set_sort_column_id (0, gtk.SORT_DESCENDING)
         self.treeview.set_model(self.store)
         self.treeview.set_rules_hint (True)
@@ -241,6 +254,8 @@ class JobViewer (GtkGUI, monitor.Watcher):
         self.treeview.connect ('popup-menu', self.on_treeview_popup_menu)
         self.treeview.connect ('cursor-changed',
                                self.on_treeview_cursor_changed)
+        self.store.connect ('row-changed',
+                            self.on_treemodel_row_changed)
 
         self.JobsWindow.set_icon_name (ICON)
         self.JobsWindow.hide ()
@@ -265,24 +280,24 @@ class JobViewer (GtkGUI, monitor.Watcher):
 
         self.statusbar_set = False
 
+        theme = gtk.icon_theme_get_default ()
+        self.icon_jobs = theme.load_icon (ICON, 22, 0)
+        self.icon_jobs_processing = theme.load_icon ("printer-printing",
+                                                     22, 0)
+        self.icon_no_jobs = self.icon_jobs.copy ()
+        self.icon_no_jobs.fill (0)
+        self.icon_jobs.composite (self.icon_no_jobs,
+                                  0, 0,
+                                  self.icon_no_jobs.get_width(),
+                                  self.icon_no_jobs.get_height(),
+                                  0, 0,
+                                  1.0, 1.0,
+                                  gtk.gdk.INTERP_BILINEAR,
+                                  127)
         if self.trayicon:
             self.statusicon = gtk.StatusIcon ()
-            theme = gtk.icon_theme_get_default ()
             pixbuf = theme.load_icon (ICON, 22, 0)
-            self.statusicon.set_from_pixbuf (pixbuf)
-            self.icon_jobs = self.statusicon.get_pixbuf ()
-            self.icon_jobs_processing = theme.load_icon ("printer-printing",
-                                                         22, 0)
-            self.icon_no_jobs = self.icon_jobs.copy ()
-            self.icon_no_jobs.fill (0)
-            self.icon_jobs.composite (self.icon_no_jobs,
-                                      0, 0,
-                                      self.icon_no_jobs.get_width(),
-                                      self.icon_no_jobs.get_height(),
-                                      0, 0,
-                                      1.0, 1.0,
-                                      gtk.gdk.INTERP_BILINEAR,
-                                      127)
+            self.statusicon.set_from_pixbuf (pixbuf)    
             self.set_statusicon_from_pixbuf (self.icon_no_jobs)
             self.statusicon.connect ('activate', self.toggle_window_display)
             self.statusicon.connect ('popup-menu', self.on_icon_popupmenu)
@@ -315,6 +330,10 @@ class JobViewer (GtkGUI, monitor.Watcher):
                     notification.close ()
                     notification.set_data ('closed', True)
 
+        if self.job_creation_times_timer != None:
+            gobject.source_remove (self.job_creation_times_timer)
+            self.job_creation_times_timer = None
+
         if self.exit_handler:
             self.exit_handler (self)
 
@@ -322,15 +341,18 @@ class JobViewer (GtkGUI, monitor.Watcher):
         self.process_pending_events = whether
 
     # Handle "special" status icon
-    def set_special_statusicon (self, iconname):
+    def set_special_statusicon (self, iconname, tooltip=None):
         self.special_status_icon = True
         self.statusicon.set_from_icon_name (iconname)
         self.set_statusicon_visibility ()
+        if tooltip != None:
+            self.set_statusicon_tooltip (tooltip=tooltip)
 
     def unset_special_statusicon (self):
         self.special_status_icon = False
         self.statusicon.set_from_pixbuf (self.saved_statusicon_pixbuf)
         self.set_statusicon_visibility ()
+        self.set_statusicon_tooltip ()
 
     def notify_new_printer (self, printer, notification):
         self.new_printer_notifications[printer] = notification
@@ -422,7 +444,7 @@ class JobViewer (GtkGUI, monitor.Watcher):
                     need_update = False
                     t = time.strftime ("%B %Y", time.localtime (created))
 
-            self.store.set_value (iter, 5, t)
+            self.store.set_value (iter, 1, t)
 
         if need_update and not self.job_creation_times_timer:
             t = gobject.timeout_add (60 * 1000, self.update_job_creation_times)
@@ -451,19 +473,30 @@ class JobViewer (GtkGUI, monitor.Watcher):
         del self.troubleshooter
 
     def add_job (self, job, data, connection=None):
+        self.update_job (job, data, connection=connection)
+
         store = self.store
         iter = self.store.append (None)
         store.set_value (iter, 0, job)
-        store.set_value (iter, 1, data.get('job-originating-user-name',
-                                           _("Unknown")))
-        store.set_value (iter, 2, data.get('job-name', _("Unknown")))
         debugprint ("Job %d added" % job)
         self.jobiters[job] = iter
-        if not self.job_creation_times_timer:
-            t = gobject.timeout_add (1000, self.update_job_creation_times)
-            self.job_creation_times_timer = t
 
-        self.update_job (job, data, connection=connection)
+        range = self.treeview.get_visible_range ()
+        if range != None:
+            (start, end) = range
+            if (self.store.get_sort_column_id () == (0,
+                                                     gtk.SORT_DESCENDING) and
+                start == (1,)):
+                # This job was added job above the visible range, and
+                # we are sorting by descending job ID.  Scroll to it.
+                self.treeview.scroll_to_cell ((0,), None, False, 0.0, 0.0)
+
+        if not self.job_creation_times_timer:
+            def start_updating_job_creation_times():
+                self.update_job_creation_times ()
+                return False
+
+            gobject.timeout_add (500, start_updating_job_creation_times)
 
     def update_job (self, job, data, connection=None):
         # Fetch required attributes for this job if they are missing.
@@ -489,20 +522,9 @@ class JobViewer (GtkGUI, monitor.Watcher):
             if attrs:
                 data.update (attrs)
 
-        store = self.store
-        iter = self.jobiters[job]
         self.jobs[job] = data
 
-        printer = data['job-printer-name']
-        store.set_value (iter, 3, printer)
-
-        size = _("Unknown")
-        if data.has_key ('job-k-octets'):
-            size = str (data['job-k-octets']) + 'k'
-        store.set_value (iter, 4, size)
-
         job_requires_auth = False
-        c = None
         try:
             jstate = data.get ('job-state', cups.IPP_JOB_PROCESSING)
             s = int (jstate)
@@ -543,72 +565,19 @@ class JobViewer (GtkGUI, monitor.Watcher):
         except cups.IPPError, (e, m):
             pass
 
-        job_requires_auth = (s == cups.IPP_JOB_HELD and
-                             data.get ('job-hold-until', 'none') ==
-                             'auth-info-required')
-        state = None
-        if job_requires_auth:
-            state = _("Held for authentication")
-        elif s == cups.IPP_JOB_HELD:
-            state = _("Held")
-            until = data.get ('job-hold-until')
-            if until != None:
-                try:
-                    colon1 = until.find (':')
-                    if colon1 != -1:
-                        now = time.gmtime ()
-                        hh = int (until[:colon1])
-                        colon2 = until[colon1 + 1:].find (':')
-                        if colon2 != -1:
-                            colon2 += colon1 + 1
-                            mm = int (until[colon1 + 1:colon2])
-                            ss = int (until[colon2 + 1:])
-                        else:
-                            mm = int (until[colon1 + 1:])
-                            ss = 0
-
-                        day = now.tm_mday
-                        if (hh < now.tm_hour or
-                            hh == now.tm_hour and
-                            (mm < now.tm_min or
-                             mm == now.tm_min and ss < now.tm_sec)):
-                            day += 1
-
-                        hold = (now.tm_year, now.tm_mon, day,
-                                hh, mm, ss, 0, 0, -1)
-                        local = time.localtime (time.mktime (hold) - time.timezone)
-                        state = _("Held until %s") % time.strftime ("%X", local)
-                except ValueError:
-                    pass
-            if until == "day-time":
-                state = _("Held until day-time")
-            elif until == "evening":
-                state = _("Held until evening")
-            elif until == "night":
-                state = _("Held until night-time")
-            elif until == "second-shift":
-                state = _("Held until second shift")
-            elif until == "third-shift":
-                state = _("Held until third shift")
-            elif until == "weekend":
-                state = _("Held until weekend")
-        else:
-            try:
-                state = { cups.IPP_JOB_PENDING: _("Pending"),
-                          cups.IPP_JOB_PROCESSING: _("Processing"),
-                          cups.IPP_JOB_STOPPED: _("Stopped"),
-                          cups.IPP_JOB_CANCELED: _("Canceled"),
-                          cups.IPP_JOB_ABORTED: _("Aborted"),
-                          cups.IPP_JOB_COMPLETED: _("Completed") }[s]
-            except IndexError:
-                pass
-
-        if state == None:
-            state = _("Unknown")
-        store.set_value (iter, 6, state)
+        # Invalidate the cached status description and redraw the treeview.
+        try:
+            del data['_status_text']
+        except KeyError:
+            pass
+        self.treeview.queue_draw ()
 
         # Check whether authentication is required.
         if self.trayicon:
+            job_requires_auth = (s == cups.IPP_JOB_HELD and
+                                 data.get ('job-hold-until', 'none') ==
+                                 'auth-info-required')
+
             if (job_requires_auth and
                 not self.auth_info_dialogs.has_key (job)):
                 try:
@@ -635,10 +604,8 @@ class JobViewer (GtkGUI, monitor.Watcher):
                     else:
                         (serverport, rest) = urllib.splithost (rest)
                         (server, port) = urllib.splitnport (serverport)
-                    username = pwd.getpwuid (os.getuid ())[0]
                     keyring_attrs.update ({ "server": str (server.lower ()),
-                                            "protocol": str (scheme),
-                                            "user": str (username)})
+                                            "protocol": str (scheme)})
 
                 if job in self.authenticated_jobs:
                     # We've already tried to authenticate this job before.
@@ -687,19 +654,9 @@ class JobViewer (GtkGUI, monitor.Watcher):
                         c._end_operation ()
                         nonfatalException ()
 
-                self.display_auth_info_dialog (job)
-
-    def on_auth_notification_closed (self, notification, reason=None):
-        job = notification.get_data ('job-id')
-        debugprint ("auth notification closed for job %s" % job)
-        self.auth_notifications[job].set_data ('closed', True)
-        del self.auth_notifications[job]
-
-    def on_auth_notification_authenticate (self, notification, action):
-        job = notification.get_data ('job-id')
-        keyring_attrs = notification.get_data ('keyring-attrs')
-        debugprint ("auth notification authenticate for job %s" % job)
-        self.display_auth_info_dialog (job, keyring_attrs)
+                username = pwd.getpwuid (os.getuid ())[0]
+                keyring_attrs["user"] = str (username)
+                self.display_auth_info_dialog (job, keyring_attrs)
 
     def display_auth_info_dialog (self, job, keyring_attrs=None):
         data = self.jobs[job]
@@ -790,8 +747,9 @@ class JobViewer (GtkGUI, monitor.Watcher):
                                            attrs.get ("protocol"))
                     ind = auth_info_required.index ('password')
                     secret = auth_info[ind]
-                    gnomekeyring.item_create_sync (keyring, type, name,
-                                                   attrs, secret, True)
+                    id = gnomekeyring.item_create_sync (keyring, type, name,
+                                                        attrs, secret, True)
+                    debugprint ("keyring: created id %d for %s" % (id, name))
             except:
                 nonfatalException ()
 
@@ -808,7 +766,6 @@ class JobViewer (GtkGUI, monitor.Watcher):
             return
 
         open_notifications = len (self.new_printer_notifications.keys ())
-        open_notifications += len (self.completed_job_notifications.keys ())
         for reason, notification in self.state_reason_notifications.iteritems():
             if notification.get_data ('closed') != True:
                 open_notifications += 1
@@ -834,6 +791,9 @@ class JobViewer (GtkGUI, monitor.Watcher):
         if event.button == 3:
             self.show_treeview_popup_menu (treeview, event, event.button)
 
+    def on_treemodel_row_changed (self, model, path, iter):
+        self.on_treeview_cursor_changed (self.treeview)
+
     def on_treeview_cursor_changed (self, treeview):
         path, column = treeview.get_cursor ()
         cancel = self.job_ui_manager.get_action ("/cancel-job")
@@ -1036,36 +996,46 @@ class JobViewer (GtkGUI, monitor.Watcher):
         return True
 
     ## Icon manipulation
-    def add_state_reason_emblem (self, pixbuf):
-        if self.worst_reason != None:
+    def add_state_reason_emblem (self, pixbuf, printer=None):
+        worst_reason = None
+        if printer == None and self.worst_reason != None:
             # Check that it's valid.
             printer = self.worst_reason.get_printer ()
             found = False
-            for reason in self.printer_state_reasons[printer]:
+            for reason in self.printer_state_reasons.get (printer, []):
                 if reason == self.worst_reason:
-                    found = True
+                    worst_reason = self.worst_reason
                     break
-            if not found:
+            if worst_reason == None:
                 self.worst_reason = None
 
-        if self.worst_reason != None:
-            level = self.worst_reason.get_level ()
+        if printer != None:
+            for reason in self.printer_state_reasons.get (printer, []):
+                if worst_reason == None:
+                    worst_reason = reason
+                elif reason > worst_reason:
+                    worst_reason = reason
+
+        if worst_reason != None:
+            level = worst_reason.get_level ()
             if level > StateReason.REPORT:
                 # Add an emblem to the icon.
                 icon = StateReason.LEVEL_ICON[level]
                 pixbuf = pixbuf.copy ()
-                theme = gtk.icon_theme_get_default ()
-
-                emblem = theme.load_icon (icon, 22, 0)
-                emblem.composite (pixbuf,
-                                  pixbuf.get_width () / 2,
-                                  pixbuf.get_height () / 2,
-                                  emblem.get_width () / 2,
-                                  emblem.get_height () / 2,
-                                  pixbuf.get_width () / 2,
-                                  pixbuf.get_height () / 2,
-                                  0.5, 0.5,
-                                  gtk.gdk.INTERP_BILINEAR, 255)
+                try:
+                    theme = gtk.icon_theme_get_default ()
+                    emblem = theme.load_icon (icon, 22, 0)
+                    emblem.composite (pixbuf,
+                                      pixbuf.get_width () / 2,
+                                      pixbuf.get_height () / 2,
+                                      emblem.get_width () / 2,
+                                      emblem.get_height () / 2,
+                                      pixbuf.get_width () / 2,
+                                      pixbuf.get_height () / 2,
+                                      0.5, 0.5,
+                                      gtk.gdk.INTERP_BILINEAR, 255)
+                except gobject.GError:
+                    debugprint ("No %s icon available" % icon)
 
         return pixbuf
 
@@ -1202,57 +1172,6 @@ class JobViewer (GtkGUI, monitor.Watcher):
         self.set_statusicon_visibility ()
         return
 
-    def notify_completed_job (self, jobid):
-        job = self.jobs.get (jobid, {})
-        document = job.get ('job-name', _("Unknown"))
-        printer_uri = job.get ('job-printer-uri')
-        if printer_uri != None:
-            # Determine if this printer is remote.  There's no need to
-            # show a notification if the printer is connected to this
-            # machine.
-
-            # Find out the device URI.  We might already have
-            # determined this if authentication was required.
-            device_uri = job.get ('device-uri')
-
-            if device_uri == None:
-                pattrs = ['device-uri']
-                c = authconn.Connection (self.JobsWindow,
-                                         host=self.host,
-                                         port=self.port,
-                                         encryption=self.encryption)
-                attrs = c.getPrinterAttributes (uri=printer_uri,
-                                                requested_attributes=pattrs)
-                device_uri = attrs.get ('device-uri')
-
-            if device_uri != None:
-                (scheme, rest) = urllib.splittype (device_uri)
-                if scheme not in ['socket', 'ipp', 'http', 'smb']:
-                    return
-
-        printer = job.get ('job-printer-name', _("Unknown"))
-        notification = pynotify.Notification (_("Job %d completed") % jobid,
-                                              _("Document `%s' has finished "
-                                                "printing on `%s'.") %
-                                              (document, printer),
-                                              'printer')
-        notification.set_urgency (pynotify.URGENCY_LOW)
-        notification.connect ('closed',
-                              self.on_completed_job_notification_closed)
-        notification.set_data ('jobid', jobid)
-        self.completed_job_notifications[jobid] = notification
-        self.set_statusicon_visibility ()
-        notification.attach_to_status_icon (self.statusicon)
-        try:
-            notification.show ()
-        except gobject.GError:
-            nonfatalException ()
-
-    def on_completed_job_notification_closed (self, notification, reason=None):
-        jobid = notification.get_data ('jobid')
-        del self.completed_job_notifications[jobid]
-        self.set_statusicon_visibility ()
-
     ## monitor.Watcher interface
     def current_printers_and_jobs (self, mon, printers, jobs):
         monitor.Watcher.current_printers_and_jobs (self, mon, printers, jobs)
@@ -1321,36 +1240,21 @@ class JobViewer (GtkGUI, monitor.Watcher):
             printer = uri
         jobdata['job-printer-name'] = printer
 
-        any_active = len (self.active_jobs) > 0
         if self.job_is_active (jobdata):
             self.active_jobs.add (jobid)
         elif jobid in self.active_jobs:
             self.active_jobs.remove (jobid)
-        if (len (self.active_jobs) > 0) != any_active:
-            self.update_status ()
 
         self.update_job (jobid, jobdata)
+        self.update_status ()
         jobdata = self.jobs[jobid]
 
-        # If the job has finished, let the user know.
-        if self.trayicon and (eventname == 'job-completed' or
-                              (eventname == 'job-state-changed' and
-                               event['job-state'] == cups.IPP_JOB_COMPLETED)):
-            reasons = event['job-state-reasons']
-            if type (reasons) != list:
-                reasons = [reasons]
-
-            canceled = False
-            for reason in reasons:
-                if reason.startswith ("job-canceled"):
-                    canceled = True
-                    break
-
-            if not canceled:
-                self.notify_completed_job (jobid)
-
         # Look out for stopped jobs.
-        if (self.trayicon and eventname == 'job-stopped' and
+        if (self.trayicon and
+            (eventname == 'job-stopped' or
+             (eventname == 'job-state-changed' and
+              event['job-state'] in [cups.IPP_JOB_STOPPED,
+                                     cups.IPP_JOB_PENDING])) and
             not jobid in self.stopped_job_prompts):
             # Why has the job stopped?  It might be due to a job error
             # of some sort, or it might be that the backend requires
@@ -1360,7 +1264,7 @@ class JobViewer (GtkGUI, monitor.Watcher):
             # update_job.
             may_be_problem = True
             jstate = jobdata['job-state']
-            if (jstate in [cups.IPP_JOB_PENDING, cups.IPP_JOB_PROCESSING] or
+            if (jstate == cups.IPP_JOB_PROCESSING or
                 (jstate == cups.IPP_JOB_HELD and
                  jobdata['job-hold-until'] == 'auth-info-required')):
                 # update_job already dealt with this.
@@ -1395,7 +1299,8 @@ class JobViewer (GtkGUI, monitor.Watcher):
                 elif notify_text.find ("filter errors") != -1:
                     message = _("There was a problem processing document `%s' "
                                 "(job %d).") % (document, jobid)
-                elif notify_text.find ("being paused") != -1:
+                elif (notify_text.find ("being paused") != -1 or
+                      jstate != cups.IPP_JOB_STOPPED):
                     may_be_problem = False
                 else:
                     # Give up and use the provided message untranslated.
@@ -1448,24 +1353,6 @@ class JobViewer (GtkGUI, monitor.Watcher):
     def job_removed (self, mon, jobid, eventname, event):
         monitor.Watcher.job_removed (self, mon, jobid, eventname, event)
 
-        # If the job has finished, let the user know.
-        if self.trayicon and (eventname == 'job-completed' or
-                              (eventname == 'job-state-changed' and
-                               event['job-state'] == cups.IPP_JOB_COMPLETED)):
-            reasons = event['job-state-reasons']
-            debugprint (reasons)
-            if type (reasons) != list:
-                reasons = [reasons]
-
-            canceled = False
-            for reason in reasons:
-                if reason.startswith ("job-canceled"):
-                    canceled = True
-                    break
-
-            if not canceled:
-                self.notify_completed_job (jobid)
-
         if self.jobiters.has_key (jobid):
             self.store.remove (self.jobiters[jobid])
             del self.jobiters[jobid]
@@ -1491,6 +1378,7 @@ class JobViewer (GtkGUI, monitor.Watcher):
         reason.user_notified = False
         l.append (reason)
         self.update_status ()
+        self.treeview.queue_draw ()
 
         if not self.trayicon:
             return
@@ -1523,6 +1411,7 @@ class JobViewer (GtkGUI, monitor.Watcher):
         del reasons[i]
 
         self.update_status ()
+        self.treeview.queue_draw ()
 
         if not self.trayicon:
             return
@@ -1591,16 +1480,157 @@ class JobViewer (GtkGUI, monitor.Watcher):
         monitor.Watcher.printer_removed (self, mon, printer)
         self.printer_uri_index.remove_printer (printer)
 
-    ## Printer status window
-    def set_printer_status_icon (self, column, cell, model, iter, *user_data):
-        level = model.get_value (iter, 0)
-        icon = StateReason.LEVEL_ICON[level]
-        theme = gtk.icon_theme_get_default ()
-        try:
-            pixbuf = theme.load_icon (icon, 22, 0)
-            cell.set_property("pixbuf", pixbuf)
-        except gobject.GError, exc:
-            pass # Couldn't load icon
+    ### Cell data functions
+    def _set_job_job_number_text (self, column, cell, model, iter, *data):
+        cell.set_property("text", str (model.get_value (iter, 0)))
 
-    def set_printer_status_name (self, column, cell, model, iter, *user_data):
-        cell.set_property("text", model.get_value (iter, 1))
+    def _set_job_user_text (self, column, cell, model, iter, *data):
+        jobid = model.get_value (iter, 0)
+        job = self.jobs[jobid]
+        cell.set_property("text", job.get ('job-originating-user-name',
+                                           _("Unknown")))
+
+    def _set_job_document_text (self, column, cell, model, iter, *data):
+        jobid = model.get_value (iter, 0)
+        job = self.jobs[jobid]
+        cell.set_property("text", job.get('job-name', _("Unknown")))
+
+    def _set_job_printer_text (self, column, cell, model, iter, *data):
+        jobid = model.get_value (iter, 0)
+        cell.set_property("text", self.jobs[jobid]['job-printer-name'])
+
+    def _set_job_size_text (self, column, cell, model, iter, *data):
+        jobid = model.get_value (iter, 0)
+        job = self.jobs[jobid]
+        size = _("Unknown")
+        if job.has_key ('job-k-octets'):
+            size = str (job['job-k-octets']) + 'k'
+        cell.set_property("text", size)
+
+    def _find_job_state_text (self, job):
+        data = self.jobs[job]
+        jstate = data.get ('job-state', cups.IPP_JOB_PROCESSING)
+        s = int (jstate)
+        job_requires_auth = (s == cups.IPP_JOB_HELD and
+                             data.get ('job-hold-until', 'none') ==
+                             'auth-info-required')
+        state = None
+        if job_requires_auth:
+            state = _("Held for authentication")
+        elif s == cups.IPP_JOB_HELD:
+            state = _("Held")
+            until = data.get ('job-hold-until')
+            if until != None:
+                try:
+                    colon1 = until.find (':')
+                    if colon1 != -1:
+                        now = time.gmtime ()
+                        hh = int (until[:colon1])
+                        colon2 = until[colon1 + 1:].find (':')
+                        if colon2 != -1:
+                            colon2 += colon1 + 1
+                            mm = int (until[colon1 + 1:colon2])
+                            ss = int (until[colon2 + 1:])
+                        else:
+                            mm = int (until[colon1 + 1:])
+                            ss = 0
+
+                        day = now.tm_mday
+                        if (hh < now.tm_hour or
+                            (hh == now.tm_hour and
+                             (mm < now.tm_min or
+                              (mm == now.tm_min and ss < now.tm_sec)))):
+                            day += 1
+
+                        hold = (now.tm_year, now.tm_mon, day,
+                                hh, mm, ss, 0, 0, -1)
+                        old_tz = os.environ.get("TZ")
+                        os.environ["TZ"] = "UTC"
+                        simpletime = time.mktime (hold)
+
+                        if old_tz == None:
+                            del os.environ["TZ"]
+                        else:
+                            os.environ["TZ"] = old_tz
+
+                        local = time.localtime (simpletime)
+                        state = _("Held until %s") % time.strftime ("%X", local)
+                except ValueError:
+                    pass
+            if until == "day-time":
+                state = _("Held until day-time")
+            elif until == "evening":
+                state = _("Held until evening")
+            elif until == "night":
+                state = _("Held until night-time")
+            elif until == "second-shift":
+                state = _("Held until second shift")
+            elif until == "third-shift":
+                state = _("Held until third shift")
+            elif until == "weekend":
+                state = _("Held until weekend")
+        else:
+            try:
+                state = { cups.IPP_JOB_PENDING: _("Pending"),
+                          cups.IPP_JOB_PROCESSING: _("Processing"),
+                          cups.IPP_JOB_STOPPED: _("Stopped"),
+                          cups.IPP_JOB_CANCELED: _("Canceled"),
+                          cups.IPP_JOB_ABORTED: _("Aborted"),
+                          cups.IPP_JOB_COMPLETED: _("Completed") }[s]
+            except IndexError:
+                pass
+
+        if state == None:
+            state = _("Unknown")
+
+        return state
+
+    def _set_job_status_icon (self, column, cell, model, iter, *data):
+        jobid = model.get_value (iter, 0)
+        data = self.jobs[jobid]
+        jstate = data.get ('job-state', cups.IPP_JOB_PROCESSING)
+        s = int (jstate)
+        if s == cups.IPP_JOB_PROCESSING:
+            icon = self.icon_jobs_processing
+        else:
+            icon = self.icon_jobs
+
+        if s == cups.IPP_JOB_HELD:
+            theme = gtk.icon_theme_get_default ()
+            emblem = theme.load_icon (gtk.STOCK_MEDIA_PAUSE, 22 / 2, 0)
+            copy = icon.copy ()
+            emblem.composite (copy, 0, 0,
+                              copy.get_width (),
+                              copy.get_height (),
+                              copy.get_width () / 2 - 1,
+                              copy.get_height () / 2 - 1,
+                              1.0, 1.0,
+                              gtk.gdk.INTERP_NEAREST, 255)
+            icon = copy
+        else:
+            # Check state reasons.
+            printer = data['job-printer-name']
+            icon = self.add_state_reason_emblem (icon, printer=printer)
+
+        cell.set_property ("pixbuf", icon)
+
+    def _set_job_status_text (self, column, cell, model, iter, *data):
+        jobid = model.get_value (iter, 0)
+        data = self.jobs[jobid]
+        try:
+            text = data['_status_text']
+        except KeyError:
+            text = self._find_job_state_text (jobid)
+            data['_status_text'] = text
+
+        printer = data['job-printer-name']
+        reasons = self.printer_state_reasons.get (printer, [])
+        if len (reasons) > 0:
+            worst_reason = reasons[0]
+            for reason in reasons[1:]:
+                if reason > worst_reason:
+                    worst_reason = reason
+            (title, unused) = worst_reason.get_description ()
+            text += " - " + title
+
+        cell.set_property ("text", text)
